#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// A logical line has a line number and some character data.  Since it may be
// composed of multiple actual lines separated by escaped newlines, we store
// the boundaries between these lines in a line_lengths array.
typedef struct {
    int number;           // the line number
    int line_lengths_sz;  // the number of physical lines in this logical line
    int *line_lengths;    // the boundaries between physical lines
    char *data;           // the character data of this logical line
} line_t;

// Free the heap allocated data in a line record (data and line_lengths)
void free_line(line_t line) {
    free(line.line_lengths);
    free(line.data);
}

// Compare two lines by their line numbers
int line_cmp_by_number(const void *a, const void *b) {
    const line_t *aa = a;
    const line_t *bb = b;
    return (aa->number - bb->number);
}

// Compare two lines by their contents (using strcmp)
int line_cmp_by_content(const void *a, const void *b) {
    const line_t *aa = a;
    const line_t *bb = b;
    return strcmp(aa->data, bb->data);
}

// Compare two lines, primarily by their contents, breaking ties by line number
int line_cmp(const void *a, const void *b) {
    int strcmp = line_cmp_by_content(a, b);
    if (strcmp == 0) {
        return line_cmp_by_number(a, b);
    } else {
        return strcmp;
    }
}

// The input text is a sequence of (logical) lines
typedef struct {
    int capacity;  // the number of lines that fit in lines
    int size;      // the number of lines currently stored in lines
    line_t *lines; // the lines in this input
} text_t;

// Allocate a text structure
text_t alloc_text() {
    text_t text;
    text.capacity = 16;
    text.size = 0;
    text.lines = malloc(text.capacity * sizeof(line_t));
    return text;
}

// Free the heap allocated data stored in this text object
void free_text(text_t text) {
    for (int i = 0; i < text.size; ++i) {
        free_line(text.lines[i]);
    }
    free(text.lines);
}

// Double the size of the lines array in this text object in order to make room
// for more lines to be added.
void text_grow(text_t *text) {
    text->capacity *= 2;
    text->lines = realloc(text->lines, text->capacity * sizeof(line_t));
}

// Add a new line to the end of this text
void text_push(text_t *text, line_t line) {
    if (text->size == text->capacity) {
        text_grow(text);
    }
    text->lines[text->size++] = line;
}

// Read the input and return a text object storing it
text_t read_input() {
    text_t text = alloc_text();
    line_t line = { 1, 1, NULL, NULL };
    int line_length;
    size_t buf_size;

    // Read input lines one by one and append them to the text
    while ((line_length = getline(&line.data, &buf_size, stdin)) >= 0) {
        if (line_length > 0 && line.data[line_length - 1] == '\n') {
            line.data[--line_length] = '\0';
        }
        line.line_lengths = malloc(sizeof(int));
        *line.line_lengths = line_length;
        text_push(&text, line);
        line.data = NULL;
        ++line.number;
    }

    // The last call of getline allocates a char array that is not added to the
    // text object.  Need to free it here.
    if (line.data) {
        free(line.data);
    }

    return text;
}

// Merge all the lines in text from line number from, inclusive, to line number
// to, exclusive, into a single logical line.
void merge_line_range(text_t *text, int from, int to) {

    // Do the merge only if there is more than one line to be merged
    if (to > from + 1) {

        // This is where we'll collect the boundaries between the merged lines.
        int *merged_line_lengths = malloc((to - from) * sizeof(int));

        // This is where we'll collect the characters of the constructed
        // logical line.  dest is a helper.
        char *merged_data, *dest;

        // The total length of all merged lines, that is, the length of the
        // logical line.
        int total_length;

        // Collect the boundaries between physical lines in this logical line
        // and calculate the total length.  Subtract 1 for each merged line
        // except the last because we will discard the trailing '\'.
        for (int i = from; i < to; ++i) {
            merged_line_lengths[i - from] = *text->lines[i].line_lengths - 1;
        }
        ++merged_line_lengths[to - from - 1];
        for (int i = total_length = 0; i < to - from; ++i) {
            total_length += merged_line_lengths[i];
        }

        // Allocate a new character array for the merged logical line and copy
        // the merged lines into it.
        merged_data = malloc((total_length + 1) * sizeof(char));
        dest = merged_data;
        for (int i = from; i < to; ++i) {
            dest = stpcpy(dest, text->lines[i].data);
            --dest; // This makes sure the trailing '\' is overwritten by the
                    // next substring.

            // Free the character data and line length arrays associated with
            // the merged lines
            free_line(text->lines[i]);
        }

        // The fromth line now gets the line_lengths array and character data
        // constructed here.
        text->lines[from].line_lengths_sz = to - from;
        text->lines[from].line_lengths = merged_line_lengths;
        text->lines[from].data = merged_data;
    }
}

// Merge line ranges joined by escaped newlines
void merge_lines(text_t *text) {
    int i;  // The first line in the line range to be merged
    int j;  // The current line
    int k;  // The number of the next merged line to be produced

    for (i = j = k = 0; j < text->size; ++j) {

        // If the current line does not end with a '\', then it's the last line
        // in a block.  Merge the lines in the current block and copy them to
        // position k.
        if (text->lines[j].data[*text->lines[j].line_lengths - 1] != '\\') {
            merge_line_range(text, i, j + 1);
            text->lines[k++] = text->lines[i];
            i = j + 1;
        }
    }

    // If i < text->size, then we have a somewhat degenerate input because the
    // last line is terminated by a backslash.  Merge the final block of lines.
    if (i < text->size) {
        merge_line_range(text, i, text->size);
        text->lines[k++] = text->lines[i];
    }

    // Record the new number of logical lines
    text->size = k;
}

// Eliminate duplicate adjacent lines
void eliminate_duplicates(text_t *text) {
    int i, j;
    for (i = 1, j = 0; i < text->size; ++i) {
        if (line_cmp_by_content(text->lines + i, text->lines + j) == 0) {
            free_line(text->lines[i]);
        } else {
            text->lines[++j] = text->lines[i];
        }
    }
    text->size = j + 1;
}

// Print a given logical line.  This uses the line_lengths array to break it up
// into the physical lines it is made up of.
void print_line(line_t line) {
    char *start = line.data;
    for (int i = 0; i < line.line_lengths_sz; start += line.line_lengths[i++]) {
        printf("%.*s", line.line_lengths[i], start);
        if (i < line.line_lengths_sz - 1) {
            putchar('\\');
        }
        putchar('\n');
    }
}

// Merge two sorted sequences.  The input is given in from.  The output is
// to be written to to.  The left half stretches from index 0 to index m-1.
// The right half stretches from index m to index m-1.
void _merge(char *from, char *to,
            size_t m, size_t n, size_t elemsz,
            int (*cmp)(const void *, const void *)) {
    char *left, *right, *mid, *end, *dst;
    mid = from + m * elemsz;
    end = from + n * elemsz;

    // Copy the next smallest input element to the output as long as we have
    // not exhausted either of the two input halves.
    for (left = from, right = mid, dst = to;
         left != mid && right != end;
         dst += elemsz) {
        if (cmp(left, right) > 0) {
            memcpy(dst, right, elemsz);
            right += elemsz;
        } else {
            memcpy(dst, left, elemsz);
            left += elemsz;
        }
    }

    // If the left still has elements, copy the remaining elements to the end
    // of the output
    if (left != mid) {
        memcpy(dst, left, mid - left);
    }

    // If the right still has elements, copy the remaining elements to the end
    // of the output
    if (right != end) {
        memcpy(dst, right, end - right);
    }
}

// The main body of the merge sort implementation.  from holds the n input
// elements to be sorted.  to is where we'd normally expect the sorted output
// to be placed, but that depends on the recursion and the resulting copying
// back and forth between from and to during merge steps.
//
// Returns from or to, depending on where the sorted output ended up.
char* _merge_sort(char *from, char *to,
                  size_t n, size_t elemsz,
                  int (*cmp)(const void *, const void *)) {

    // We need to do something only if the input has at least 2 elements.
    if (n > 1) {

        // Recursively sort the left and right halves
        size_t m = n / 2;
        char *left = _merge_sort(from, to, m, elemsz, cmp);
        char *right = _merge_sort(from + m * elemsz, to + m * elemsz,
                                  n - m, elemsz, cmp);

        // Make sure the left and right halves are both in from or both in to.
        // This may not be the case because the two halves may differ in size
        // by one element, which may lead to different recursion depths of
        // these two calls.  We always copy the left half because it is
        // guaranteed to be no bigger than the right half.
        if (right != left + m * elemsz) {
            if (left == from) {
                memcpy(to, left, m * elemsz);
                left = to;
            } else {
                memcpy(from, left, m * elemsz);
                left = from;
            }
        }

        // Now merge the two sorted halves from from to to or from to to from
        // depending on where they are currently stored.
        if (left == from) {
            _merge(from, to, m, n, elemsz, cmp);
            return to;
        } else {
            _merge(to, from, m, n, elemsz, cmp);
            return from;
        }
    } else {
        return from;
    }
}

// The public interface of merge_sort.  The real work is done by the
// _merge_sort worker above.  The reason this is split into two functions is
// that it allows us to cut the amount of copying roughly in half.  Instead of
// copying the input into a helper space and then merging it back into the
// input space in each merge step, we allocate a helper space here and swap the
// roles of input space and helper space in each merge step.  As a consequence,
// the final sorted output may end up in the helper space instead of the input
// space.  If that's the case, we perform one final copy from the helper space
// back to the input.
//
// A possibly even greater benefit of this approach is that we allocate one
// block of space here and perform no further memory allocations while Merge
// Sort runs.  The "normal" strategy would force us to allocate a helper array
// in each merge step.  Since there are n - 1 such merge steps overall, this is
// potentially very costly.
void merge_sort(void *base, size_t n, size_t elemsz,
                int (*cmp)(const void *, const void *)) {
    void *tmp = malloc(n * elemsz);
    void *result = _merge_sort(base, tmp, n, elemsz, cmp);
    if (result == tmp) {
        memcpy(base, result, n * elemsz);
    }
    free(tmp);
}

// Print a usage message and exit
void usage(char *progname) {
    fprintf(stderr, "USAGE: %s [-m]\n", progname);
    exit(1);
}

// The main function
int main(int argc, char **argv) {

    // If there is more than one command line argument or the only command line
    // argument is not "-m", then that's not accepted.
    if (argc > 2 || (argc == 2 && strcmp(argv[1], "-m"))) {
        usage(argv[0]);
    }

    // Read the input
    text_t text = read_input();

    // Merge lines separated by escaped newlines if the "-m" flag was given
    if (argc == 2) {
        merge_lines(&text);
    }

    // Sort the lines by their content
    merge_sort(text.lines, text.size, sizeof(line_t), line_cmp);

    // Eliminate duplicate adjacent lines
    eliminate_duplicates(&text);

    // Arrange the remaining lines back in their input order
    merge_sort(text.lines, text.size, sizeof(line_t), line_cmp_by_number);

    // Print the lines
    for (int i = 0; i < text.size; ++i) {
        print_line(text.lines[i]);
    }

    // Release the memory used by the text object
    free_text(text);

    return 0;
}
